from sklearn.cluster import dbscan

################################################################################
#
# Reductions and clustering
#
################################################################################

def clusterDistMatrix( anchorTable, args):
    kmerAssignments = {}    
    iam, rowNames = anchorTable.asInverseAdjacency(crackThresh=args.crack_thresh)

    _, labels = dbscan( iam,
                        eps=args.dbscan_eps,
                        min_samples=args.dbscan_min_samples,
                        metric='precomputed')
    
    for ind, kmer in rowNames.items():
        clust = labels[ind]
        if clust == -1:
            continue
        kmerAssignments[kmer] = clust

    if args.rescue_unassigned:
        rescueUnassigned( kmerAssignments, anchorTable, args)
        
    return kmerAssignments


def rescueUnassigned(assignments, anchorTable, args):
    adjacency = anchorTable.asAdjacencyList()
    unassigned = [kmer for kmer in adjacency.keys() if kmer not in assignments]

    clusters = {}
    for kmer in unassigned:
        for neib in adjacency[kmer]:
            try:
                clust = assignments[neib]
            except KeyError:
                continue

            try:
                clusters[kmer][clust] += 1
            except KeyError:
                try:
                    clusters[kmer][clust] = 1
                except KeyError:
                    clusters[kmer] = { clust : 1}

    for kmer, cls in clusters.items():
        potentials = []
        for cl, count in cls.items():
            if count >= args.min_rescue:
                potentials.append(cl)
        if len(potentials) == 1:
            assignments[kmer] = potentials[0]


            
            
    



import sys
import click
from .kmers import MinSparseKmerSet

################################################################################

class BarcodeKmerTable:

    def __init__(self, barcode):
        self.barcode = barcode
        self.nreads = 0
        self.readKmerSets = {}
        self.reverseMap = None
        self.kmersWithCounts = {}

    def addKmersFromRead(self, readId, kmerSet):
        self.nreads += 1
        self.readKmerSets[readId] = kmerSet
        for kmer in kmerSet:
            try:
                self.kmersWithCounts[kmer] += 1
            except KeyError:
                self.kmersWithCounts[kmer] = 1
                    
    def _buildReverseMap(self):
        self.reverseMap = {}
        for rp, kmerSet in self.readKmerSets.items():
            for kmer in kmerSet:
                try:
                    self.reverseMap[kmer].append(rp)
                except KeyError:
                    self.reverseMap[kmer] = [rp]

    def kmerToReadPairs(self, kmer):
        if self.reverseMap is None:
            self._buildReverseMap()
        return self.reverseMap[kmer]

    def allKmers(self):
        # it is important that kmers that occur multiple times are returned multiple times
        for kmer, count in self.kmersWithCounts.items():
            for _ in range(count):
                yield kmer

    def containsKmer(self, kmer):
        return kmer in self.kmersWithCounts

    def __iter__(self):
        return iter(self.allKmers())

    def __len__(self):
        return self.nreads
    
################################################################################
    
def iterBarcodesAsDocuments(filelike, k, w, dropout, verbose=False):
    for nBC, bc in enumerate( iterReadCloud(filelike)):
        if verbose and ((nBC % 100) == 0):
            sys.stderr.write('\rparsed {:,} barcodes'.format(nBC))
        if len(bc) < args.dropout:
            continue
        
        bcKmers = BarcodeKmerTable(bc.barcode)
        for rp in bc:
            kmerSet = MinSparseKmerSet( k, w, [rp.r1.seq, rp.r2.seq])
            bcKmers.addKmersFromRead(rp.sid, kmerSet)
        yield bcKmers
            
    if verbose:
        sys.stderr.write('\rparsed {:,} barcodes\n'.format(nBC))


@click.command()
@click.option('-k', default=20, help='Lengths of kmers')
@click.option('-w', default=40, help='Length of window (in bases, not kmers)')
@click.option('-d', '--dropout', default=100, help='Ignore smaller read clouds')
@click.option('-v/-q', '--verbose/--quiet', default=True, help='Periodically report output')
def printBarcodesAsDocuments(k, w, dropout, verbose):
    for bc in iterBarcodesAsDocuments(sys.stdin, k, w, dropout, verbose=verbose):
        sys.stdout.write(bc.barcode)
        for kmer in bc:
            sys.stdout.write('\t{}'.format(kmer))
        sys.stdout.write('\n')

################################################################################

def findStopKmers(filelike, k, w, dropout, meanMultiplier=10, verbose=False):
    kmerCounts = {}
    totalCount = 0
    for nBC, bc in enumerate( iterReadCloud(filelike)):
        if verbose and ((nBC % 100) == 0):
            sys.stderr.write('\rparsed {:,} barcodes'.format(nBC))
        if len(bc) < args.dropout:
            continue
        for rp in bc:
            kmerSet = MinSparseKmerSet( k, w, [rp.r1.seq, rp.r2.seq])
            for kmer in kmerSet:
                totalCount += 1
                try:
                    kmerCounts[kmer] += 1
                except KeyError:
                    kmerCounts[kmer] = 1
    if verbose:
        sys.stderr.write('\rparsed {:,} barcodes\n'.format(nBC))

    aveCount = totalCount / len(kmerCounts)
    cutoff = aveCount * meanMultiplier:
    for kmer, count in kmerCoutns.items():
        if count > cutoff:
            yield kmer

            
        
@click.command()
@click.option('-k', default=20, help='Lengths of kmers')
@click.option('-w', default=40, help='Length of window (in bases, not kmers)')
@click.option('-m', default=10, help='Find kmers that are more than Nx as common as average')
@click.option('-d', '--dropout', default=100, help='Ignore smaller read clouds')
@click.option('-v/-q', '--verbose/--quiet', default=True, help='Periodically report output')
def printStopKmers(k, w, m, dropout, verbose):
    for kmer in findStopKmers(sys.stdin, k, w, dropout,
                            meanMultiplier=m,
                            verbose=verbose):
        print(kmer)

            
if __name__ == '__main__':
    printBarcodesAsDocuments()



import click
from math import log2
from sys import stderr, stdin


def aln_purity(aln_counts, size):
    return max(aln_counts) / size


def aln_entropy(aln_counts, size):
    H = 0
    for count in aln_counts:
        p = count / size
        H += p * log2(p)
    if H < 0:
        H = -H
    return H


def eval_cluster(alns):
    size = len(alns) / 2
    counts = {}
    for aln in alns:
        try:
            counts[aln] += 0.5
        except KeyError:
            counts[aln] = 0.5
    aln_counts = list(counts.values())
    return size, aln_purity(aln_counts, size), aln_entropy(aln_counts, size)


def parse_annotated_fastq(fqf, count_all=False):
    for i, line in enumerate(fqf):
        if (i % 4) == 0:
            tkns = line.strip().split()
            al_token = None
            bc_token = None
            for tkn in tkns:
                if ('AL' in tkn) and ('UNAL' not in tkn):
                    al_token = tkn.split(':')[1]
                if ('BX' in tkn) and (count_all or 'E:' in tkn):
                    bc_token = tkn
                    if count_all:
                        bc_token = bc_token.split('E:')[0]
            if al_token and bc_token:
                yield bc_token, al_token

                
def parse_clusters(fastq_filename, count_all=False):
    current_ebx = None
    current_alns = []
    for ebx, aln in parse_annotated_fastq(fastq_filename, count_all=count_all):
        if ebx == current_ebx:
            current_alns.append(aln)
        else:
            if current_ebx:
                yield current_alns
            current_alns = [aln]
            current_ebx = ebx
    if current_ebx:
        yield current_alns


@click.command()
@click.option('--count-all/--count-enhanced', default=False)
def main(count_all):
    total_size, total_purity, total_entropy, N = 0, 0, 0, 0
    print('size,purity,entropy')
    for cluster in parse_clusters(stdin, count_all=count_all):
        size, purity, entropy = eval_cluster(cluster)
        print(f'{size},{purity},{entropy}')
        total_size += size
        total_purity += purity
        total_entropy += entropy
        N += 1
    ave_size = total_size / N
    ave_purity = total_purity / N
    ave_entropy = total_entropy / N
    print('Ave Size, Ave Purity, Ave Entropy, N', file=stderr)
    print(f'{ave_size}, {ave_purity}, {ave_entropy}, {N}', file=stderr)
    



if __name__ == '__main__':
    main()



import sys
import argparse as ap
from .barcode_kmer_table import parseBarcodes, parseBarcodesAndRemoveStopKmers
from .build_and_filter_table import buildAndFilterTable
from .cluster_matrix import clusterDistMatrix
from .progress_bar import ProgressBar

################################################################################
#
# MAIN
#
################################################################################

def main():
    args = parseArgs()

    if args.remove_stop_kmers:
        barcodeTables = parseBarcodesAndRemoveStopKmers( sys.stdin,
                                                         args.K,
                                                         args.W,
                                                         args.dropout,
                                                         verbose=True)
    else:
        barcodeTables = parseBarcodes( sys.stdin, args.K, args.W, args.dropout,
                                       verbose=True)
    msg = '{:,} barcodes were at or above dropout threshold'
    print(msg.format(len(barcodeTables)), file=sys.stderr)

    totalAnchors = 0
    for bcTbl in barcodeTables:
        if bcTbl.numReads() >= args.anchor_dropout:
            totalAnchors += 1
    msg = '{:,} barcodes are anchors'
    print(msg.format(totalAnchors), file=sys.stderr)
    if totalAnchors == 0:
        print('No anchors found', file=sys.stderr)
        return
    progressBar = ProgressBar(totalAnchors)
    sys.stderr.write('\n')    
    progressBar.write()

    output_file = sys.stdout
    if args.output_file != '-':
        output_file = open(args.output_file, 'w')
    anchors = (bcTbl for bcTbl in barcodeTables if bcTbl.numReads() >= args.anchor_dropout)
    for  anchorTable in anchors:
        # build  and filter table
        sys.stderr.write( anchorTable.barcode)
        anchorTable = buildAndFilterTable(anchorTable, barcodeTables, args)
        if anchorTable is None: # table was too small after filtering
            #            progressBar.increment()
            sys.stderr.write('\n')
            continue

        # reduce and cluster the rows
        readAssignments = clusterDistMatrix( anchorTable, args)
        writeClusters( anchorTable, readAssignments, output_file, args)
        sys.stderr.write('\n')
        #       progressBar.increment()
    if output_file != sys.stdout:
        output_file.close()
        
def writeClusters(anchorTable, readAssignments, output_file, args):
    for readId, clustNum in readAssignments.items():
        msg = '{}\t{}\t{}\n'.format(anchorTable.barcode, readId, clustNum)
        output_file.write(msg)


################################################################################
#
# ARGS
#
################################################################################
        
def  parseArgs():
    parser = ap.ArgumentParser()

    parser.add_argument('-k', '--kmer-lens', dest='K',  default=20, type=int,
                        help='Lengths of kmers')
    parser.add_argument('-w', '--window-len', dest='W', default=40, type=int,
                        help='Window for sparse kmers')
    
    parser.add_argument('-d', '--dropout', dest='dropout', default=10, type=int,
                        help='Ignore barcodes with fewer reads')
    parser.add_argument('-a', '--anchor-dropout', dest='anchor_dropout', default=50, type=int,
                        help='Do not process anchors with fewer reads')

    parser.add_argument('--min-kmer', dest='rp_low_filter', default=1, type=float,
                        help='Filter kmers that rarely occur in other barcodes')
    parser.add_argument('--min-kmer-read', dest='min_kmer_per_read', default=1, type=float,
                        help='Require reads to have multiple kmers to overlap')
    parser.add_argument( '--max-kmer', dest='rp_high_filter', default=0.03, type=float,
                         help='Filter kmers that occur constantly in other barcodes')    
    parser.add_argument('--min-barcode', dest='bc_low_filter', default=2, type=float,
                        help='Filter barcodes with low kmer overlap')
    parser.add_argument('--max-barcode', dest='bc_high_filter', default=0.9, type=float,
                        help='Filter barcodes with high kmer overlap')    
    
    parser.add_argument('--min-rows', dest='min_rows', default=5, type=int,
                        help='Do not process tables with fewer rows (kmers)')    
    parser.add_argument('--min-cols', dest='min_cols', default=3, type=int,
                        help='Do not process tables with fewer cols (barcodes)')

    parser.add_argument('--eps',dest='dbscan_eps', default=0.26, type=float,
                        help='Distance threshold for DBSCAN clustering')
    parser.add_argument('--min-samples',dest='dbscan_min_samples',default=2, type=int,
                        help='Minimum samples in a cluster for dbscan')    

    parser.add_argument('--output',dest='output_file',default='-', type=str,
                        help='File where results should be written to. Defaults to stdout.')    


    # experimental args
    parser.add_argument('--remove-stopwords',dest='remove_stop_kmers', action='store_true', 
                        help='Remove all kmers that occur 10x more often than average')    


    parser.add_argument('--rescue-unassigned',dest='rescue_unassigned', action='store_true', 
                        help='Try to assign reads not assigned by DBSCAN')    

    parser.add_argument('--min-rescue', dest='min_rescue', default=2, type=int,
                        help='Rescue a read if it has this many connections to at most one cluster')

    parser.add_argument('--crack-edges', dest='crack_thresh', default=1, type=int,
                        help='Require that nodes that share a single edge share neighbours')

    
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    main()



from sys import stdout
import pandas as pd
import click
from sys import stderr

NRANKS = 20


def parse_minerva_file(minerva_file, read_to_taxa):
    bx_tbl, ebx_tbl = {}, {}
    with open(minerva_file) as mf:
        for line in mf:
            tkns = line.strip().split('\t')
            bx = tkns[0]
            ebx = tkns[0] + '$' + tkns[2]
            try:
                taxa = read_to_taxa[tkns[1]]
            except KeyError:
                taxa = ['unclassified']
            if len(taxa) > 8:
                taxa = taxa[:8]
            try:
                bx_tbl[bx].append(taxa)
            except KeyError:
                bx_tbl[bx] = [taxa]
            try:
                ebx_tbl[ebx].append(taxa)
            except KeyError:
                ebx_tbl[ebx] = [taxa]
    return bx_tbl, ebx_tbl


def parse_kraken_file(kraken_file):
    read_to_taxa = {}
    with open(kraken_file) as kf:
        for line in kf:
            tkns = line.strip().split('\t')
            read = tkns[0]
            taxa = tkns[1].split(';')
            read_to_taxa[read] = taxa
    return read_to_taxa


def promote(taxas, taxa_tree=None):
    if taxa_tree is None:
        taxa_tree = build_taxa_tree(taxas)
    promoted_taxas = []
    for taxa in taxas:
        root = taxa_tree
        promoted = []
        for rank in taxa:
            root = root[rank]
            promoted.append(rank)
        while len(root) == 1:
            rank = list(root.keys())[0]
            root = root[rank]
            promoted.append(rank)
        promoted_taxas.append(promoted)
    return promoted_taxas
    

def build_taxa_tree(taxas):
    taxa_tree = {}
    for taxa in taxas:
        root = taxa_tree
        for rank in taxa:
            try:
                root = root[rank]
            except KeyError:
                root[rank] = {}
                root = root[rank]
    return taxa_tree
    

def promote_and_count(bx_tbl):
    before, after  = {}, {}
    for taxas in bx_tbl.values():
        promoted = promote(taxas)
        for taxa in taxas:
            try:
                before[';'.join(taxa)] += 1
            except KeyError:
                before[';'.join(taxa)] = 1
        for taxa in promoted:
            try:
                after[';'.join(taxa)] += 1
            except KeyError:
                after[';'.join(taxa)] = 1
    return before, after


def promote_and_count2_enhanced(bx_tbl, ebx_tbl):
    tbl = {}
    for ebx, taxas in ebx_tbl.items():
        promoted_once = promote(taxas)
        bx_tree = build_taxa_tree(bx_tbl[ebx.split('$')[0]])
        promoted_twice = promote(promoted_once, bx_tree)

        for original_taxa, promoted_taxa in zip(taxas, promoted_twice):
            otaxa, ptaxa = ';'.join(original_taxa), ';'.join(promoted_taxa)
            key = otaxa + ' -> ' + ptaxa 
            try:
                tbl[key] += 1
            except KeyError:
                tbl[key] = 1
    return tbl


def promote_and_count2(bx_tbl):
    tbl = {}
    for taxas in bx_tbl.values():
        promoted = promote(taxas)
        for original_taxa, promoted_taxa in zip(taxas, promoted):
            otaxa, ptaxa = ';'.join(original_taxa), ';'.join(promoted_taxa)
            key = otaxa + ' -> ' + ptaxa 
            try:
                tbl[key] += 1
            except KeyError:
                tbl[key] = 1
    return tbl




@click.command()
@click.argument('minerva_file')
@click.argument('kraken_file')
def main(minerva_file, kraken_file):
    read_to_taxa = parse_kraken_file(kraken_file)
    bx_tbl, ebx_tbl = parse_minerva_file(minerva_file, read_to_taxa)
    '''
    inp_counts, bx_counts = promote_and_count(bx_tbl)
    _, ebx_counts = promote_and_count(ebx_tbl)
    counts = pd.DataFrame.from_dict({'inputs': inp_counts, 'standard': bx_counts, 'enhanced': ebx_counts}, orient='columns')
    '''
    counts = pd.DataFrame.from_dict({
        'standard': promote_and_count2(bx_tbl),
        'enhanced': promote_and_count2_enhanced(bx_tbl, ebx_tbl),
    }, orient='columns')
    stdout.write(counts.to_csv())


if __name__ == '__main__':
    main()



from .utils import passesFilter
from collections import Counter
import sys


################################################################################
#
# Build Table
#
################################################################################

def buildAndFilterTable( anchorTable, barcodeTables, args):
    # build and filter table
    buildTable(anchorTable, barcodeTables, args)
    if anchorTable.numColumns() == 0:
        sys.stderr.write('\tempty_table_before_filter')
        return None
    filterTable( anchorTable, args)

    nrow, ncol = anchorTable.shape()
    tableTooSmall = nrow < args.min_rows or ncol < args.min_cols
    tableIsEmpty = nrow ==0 or ncol == 0

    if not tableIsEmpty:
        sys.stderr.write(' -> RF:' + str(anchorTable.shape()))
    else:
        sys.stderr.write('\t empty_table')

    if tableTooSmall or tableIsEmpty:
        return None

    return anchorTable


def buildTable(anchorTable, otherTables, args):
    for otherTable in otherTables:
        olap = len(otherTable.kmerSet() & anchorTable.kmerSet())
        maxOlap = anchorTable.numReads()
        if passesFilter( olap,
                         maxOlap,
                         args.bc_low_filter,
                         args.bc_high_filter):
            
            if not anchorTable.hasColumn( otherTable):
                buildNewColumns(anchorTable, otherTable, args)
            
def buildNewColumns(anchorTable, otherTable, args):
    kmerCounts = {}
    for anchorKmer in anchorTable.allKmers():
        if otherTable.containsKmer(anchorKmer):
            try:
                kmerCounts[anchorKmer] += 1
            except KeyError:
                kmerCounts[anchorKmer] = 1
    validKmers = [kmer for kmer, count in kmerCounts.items() if count == 1]
    anchorReadPairs = {}
    for kmer in validKmers:
        rp = anchorTable.kmerToReadPairs(kmer)[0] 
        try:
            anchorReadPairs[rp] += 1
        except KeyError:
            anchorReadPairs[rp] = 1
    anchorReadPairs = [rp for rp, count in anchorReadPairs.items() if count >= args.min_kmer_per_read]
    anchorTable.setColumn(otherTable.barcode, anchorReadPairs)
    if otherTable.numReads() >= args.anchor_dropout:
        otherReadPairs = {}
        for kmer in validKmers:
            rp = otherTable.kmerToReadPairs(kmer)[0] 
            try:
                otherReadPairs[rp] += 1
            except KeyError:
                otherReadPairs[rp] = 1
        otherReadPairs = [rp for rp, count in otherReadPairs.items() if count >= args.min_kmer_per_read]

        otherTable.setColumn(anchorTable.barcode, otherReadPairs)
        
def filterTable(anchorTbl, args):
    sys.stderr.write(' RU:({}, {})'.format(anchorTbl.numReads(), anchorTbl.numColumns()))    
    
    filtBarcodeTbl = {}
    for barcode, readPairs in anchorTbl.asDict().items():
        numReadPairs = len(readPairs)
        if passesFilter( numReadPairs,
                         anchorTbl.numReads(),
                         args.bc_low_filter,
                         args.bc_high_filter):
            filtBarcodeTbl[barcode] = readPairs

    rpCounts = Counter()
    nbc  = 0
    for barcode, readPairs in filtBarcodeTbl.items():
        nbc += 1
        rpCounts.update(readPairs)

    sys.stderr.write(' -> bcFilt:({}, {})'.format(len(rpCounts), nbc))    
        
    goodReadPairs = set()
    for rpSID, numBarcodes in rpCounts.items():
        if passesFilter( numBarcodes,
                         len(filtBarcodeTbl),
                         args.rp_low_filter,
                         args.rp_high_filter):
            goodReadPairs.add(rpSID)
        
    filtTable = {}
    for barcode, readPairs in filtBarcodeTbl.items():
        filtRps = []
        for rpSID in readPairs:
            if rpSID in goodReadPairs:
                filtRps.append(rpSID)
        filtTable[barcode] = filtRps

    anchorTbl.tbl = filtTable




import cseqs
from .seqs import *
import click

################################################################################

class MinSparseKmerSet:
    '''
    Represents the kmers in a read
    '''
    def __init__(self, K, W, seqs, canonical=True):
        self.K = K
        self.W = W
        self.nseqs = len(seqs)
        self.makeKmers(seqs, canonical)
        
    def makeKmers(self, seqs, canonical):
        self.kmers = {}
        for seq in seqs:
            if canonical:
                kmers = cseqs.makeCanonicalKmers(seq, len(seq), self.K)
            else:
                kmers = cseqs.makeKmers(seq, len(seq), self.K)
            kmers = [(kmer, hash(kmer)) for kmer in kmers]
            kmersInWindow = self.W - self.K + 1
            numWindows =  len(kmers) - kmersInWindow + 1
            for windowStart in range( numWindows):
                window = kmers[windowStart:windowStart + kmersInWindow]

                if windowStart == 0:
                    # first window, minimum does not exist yet
                    minKmer = min(window,key=lambda x: x[1])
                    self.kmers[minKmer[0]] = 1 
                elif kmers[windowStart - 1][1] == minKmer[1]:
                    # the minimum has dropped out of the window
                    minKmer = min(window,key=lambda x: x[1])
                    self.kmers[minKmer[0]] = 1                     
                elif window[-1][1] < minKmer[1]:
                    # the new kmer is lower
                    minKmer = window[-1]
                    self.kmers[minKmer[0]] = 1                     
                    
                
                
                
    def getCount(self, kmer):
        try:
            return self.kmers[kmer]
        except KeyError:
            return 0

    def overlap(self, other):
        'Return a list of kmers that occur in both sets'
        
        assert type(other) == type(self)
        out = []
        for kmer in other:
            if kmer in self:
                out.append(kmer)
        return out

    def remove(self, kmer):
        del self.kmers[kmer]

    def removeKmers(self, stopKmers):
        self.kmers = {k:v for k,v in self.kmers.items() if k not in stopKmers}
        
    def __contains__(self, kmer):
        return kmer in self.kmers

    def __iter__(self):
        return iter(self.kmers.keys())

    def withCounts(self):
        return self.kmers.items()
    
    def __str__(self):
        out = sorted( self.kmers.keys())
        return ' '.join(out)

    def __len__(self):
        return len(self.kmers)




def makeKmers(seq, K, canon=True):
    out = []
    for start in range(len(seq)-K+1):
        kmer = seq[start:start+K]
        if canon:
            kmer = canonical(kmer)
        out.append(kmer)
    
    return out



from .fastx import *

################################################################################

class GemcodeMismatchException( Exception):
    pass

class NoGemcodeException( Exception):
    pass

class ReadCloudException(Exception):
    pass

class ChromiumReadPair( ReadPair):

    __slots__ = ('barcode',)

    def __init__(self, r1, r2):
        super( ChromiumReadPair, self).__init__(r1,r2)
        self.barcode = None
        for tag in r1.tags:
            if 'BX:' in tag:
                self.barcode = tag
                break
        if self.barcode is None:
            raise NoGemcodeException()
        for tag in r2.tags:
            if 'BX:' in tag:
                if tag != self.barcode:
                    raise GemcodeMismatchException()
    @classmethod
    def fromReadPair(cls, readPair):
        return ChromiumReadPair( readPair.r1, readPair.r2)


class ReadCloud:

    __slots__ = ('barcode', 'readPairs')
    
    def __init__(self, barcode, readPairs=[]):
        self.barcode = barcode
        self.readPairs = []
        
        for rP in readPairs:
            self.addPair(rP)
            
    def addPair(self, cRP):
        if type(cRP) != ChromiumReadPair:
            raise TypeError()
        if self.barcode is not None:
            if self.barcode != cRP.barcode:
                raise GemcodeMismatchException()
        else:
            self.barcode = cRP.barcode
        self.readPairs.append(cRP)

    def getSeqs(self):
        out = []
        for rP in self.readPairs:
            out.append(rP.r1.seq)
            out.append(rP.r2.seq)
        return out
        
    def __iter__(self):
        return iter(self.readPairs)

    def __str__(self):
        out = ''
        for rp in self.readPairs:
            out += str(rp) + '\n'
        return out

    def __len__(self):
        return len(self.readPairs)
    
################################################################################

def iterReadClouds( filelike):
    rC = ReadCloud(None)
    for rP in iterFastq( filelike, interleaved=True):
        try:
            cRP = ChromiumReadPair.fromReadPair(rP)
        except  NoGemcodeException as nge:
            continue

        try:
            rC.addPair(cRP)
        except GemcodeMismatchException as gme:
            yield rC
            rC = ReadCloud( cRP.barcode)
            rC.addPair(cRP)
    yield rC
    



from minerva.gimmebio.kmers import MinSparseKmerSet
from minerva.gimmebio.readclouds import iterReadClouds
from itertools import combinations
import numpy as np
import sys

################################################################################

def parseBarcodesAndRemoveStopKmers(filelike, K, W, dropout,
                                    meanMultiplier=10,
                                    verbose=False):
    bcTbls = parseBarcodes(filelike, K, W, dropout, verbose=verbose)
    kmerCounts = {}
    totalCount = 0
    for bcTbl in bcTbls:
        for kmer in bcTbl.allKmers():
            totalCount += 1
            try:
                kmerCounts[kmer] += 1
            except KeyError:
                kmerCounts[kmer] = 1
    aveCount = totalCount / len(kmerCounts)
    cutoff = aveCount * meanMultiplier
    stopKmers = set()
    for kmer, count in kmerCounts.items():
        if (count == 1) or (count >  cutoff):
            stopKmers.add(kmer)
    print('Removing {:,} stop and singleton kmers'.format( len(stopKmers)), file=sys.stderr)
    for bcTbl in bcTbls:
        bcTbl.removeKmers( stopKmers)
    print('Removed stop and singleton kmers', file=sys.stderr)
    return bcTbls

def parseBarcodes(filelike, K, W, dropout, verbose=False):
    bcTbls = []
    for nBC, bc in enumerate( iterReadClouds(filelike)):
        if verbose and ((nBC % 100) == 0):
            sys.stderr.write('\rparsed {:,} barcodes'.format(nBC))
        if len(bc) < dropout: 
            continue
        readKmerSets = {}
        for rp in bc:
            seqs = [rp.r1.seq, rp.r2.seq]
            readKmerSets[rp.sid] = MinSparseKmerSet( K, W, seqs)
        bcTbls.append( BarcodeTable(bc.barcode, readKmerSets))
    if verbose:
        sys.stderr.write('\rparsed {:,} barcodes\n'.format(nBC))
    return bcTbls

        
################################################################################


class BarcodeTable:

    def __init__(self, barcode, readKmerSets):
        self.barcode = barcode
        self.readKmerSets = readKmerSets
        self.reverseMap = None
        self.tbl = {}
        self._allKmers = set()
        for rp, kmerSet in self.readKmerSets.items():
            for kmer in kmerSet:
                self._allKmers.add(kmer)

    def _buildReverseMap(self):
        self.reverseMap = {}
        for rp, kmerSet in self.readKmerSets.items():
            for kmer in kmerSet:
                try:
                    self.reverseMap[kmer].append(rp)
                except KeyError:
                    self.reverseMap[kmer] = [rp]

    def removeKmers(self, stopKmers):
        assert self.reverseMap is None
        self._allKmers = {k for k in self._allKmers if k not in stopKmers}
        for rp, kmerSet in self.readKmerSets.items():
            kmerSet.removeKmers( stopKmers)
        
                   
    ############################################################################


    def hasColumn(self, otherTbl):
        return (otherTbl.barcode in self.tbl) or (otherTbl.barcode == self.barcode)

    def kmerToReadPairs(self, kmer):
        if not self.reverseMap:
            self._buildReverseMap()
        return self.reverseMap[kmer]

    def setColumn(self, barcode, readPairs):
        self.tbl[barcode] = readPairs

    def kmerSet(self):
        return self._allKmers
        
    def allKmers(self):
        # it is important that kmers that occur multiple times are returned multiple times
        for kmerSet in self.readKmerSets.values():
            for kmer in kmerSet:
                yield kmer

    def containsKmer(self, kmer):
        return kmer in self._allKmers


    def asAdjacencyList(self):
        rpInds = {}
        rowNames = {}
        for bc, rps in self.tbl.items():
            for rp in rps:
                if rp not in rpInds:
                    n = len(rpInds)
                    rpInds[rp] = n
                    rowNames[n] = rp
                    
        adjList = {}
        for bc, rps in self.tbl.items():
            for r1, r2 in combinations(rps, 2):
                try:
                    adjList[r1].add(r2)
                except KeyError:
                    adjList[r1] = set([r2])

                try:
                    adjList[r2].add(r1)
                except KeyError:
                    adjList[r2] = set([r1])
        return adjList

    
    def asInverseAdjacency(self, crackThresh=-1):
        rpInds = {}
        rowNames = {}
        for bc, rps in self.tbl.items():
            for rp in rps:
                if rp not in rpInds:
                    n = len(rpInds)
                    rpInds[rp] = n
                    rowNames[n] = rp
                    
        ws = {}
        for bc, rps in self.tbl.items():
            for r1, r2 in combinations(rps, 2):
                key = tuple( sorted( [r1,r2]))
                try:
                    ws[key] += 1
                except KeyError:
                    ws[key] = 1

        if crackThresh >= 1:
            ws = self._crackEdges(ws, crackThresh)
                    
        iam = (1000*1000) * np.ones( (len(rpInds), len(rpInds)))
        for (r1,r2), w in ws.items():
            i1 = rpInds[r1]
            i2 = rpInds[r2]
            iam[i1,i2] = 1.0 / w
            iam[i2,i1] = 1.0 / w            

        sys.stderr.write(' -> INV_ADJ:' + str(iam.shape))                    
        return iam, rowNames


    def _crackEdges(self, ws, crackThresh):
        out = {}
        adjList = self.asAdjacencyList()
        for (r1,r2), w in ws.items():
            if w == 1:
                sharedNeibs = adjList[r1] & adjList[r2]
                if len(sharedNeibs) < crackThresh:
                    continue
            out[(r1,r2)] = w
        return out
    
    def asDataFrame(self):
        df = {}
        for bc, rps in self.tbl.items():
            df[bc] = {rp:1 for rp in rps}
        df = pd.DataFrame(df)
        df.fillna(value=0, inplace=True)
        return df

    def asDict(self):
        return self.tbl

    def numColumns(self):
        return len(self.tbl)

    def numReads(self):
        return len(self.readKmerSets)

    def numRows(self):
        uniqueRows = set()
        for rps in self.tbl.values():
            for rp in rps:
                uniqueRows.add(rp)
        return len(uniqueRows)

    def shape(self):
        return self.numRows(), self.numColumns()



import click
from sys import stdin, stdout, stderr


def get_bc_map(bc_tbl):
    out = {}
    with open(bc_tbl) as bct:
        for line in bct:
            tkns = line.strip().split()
            if len(tkns) != 3:
                continue
            bc, rid, cl = tkns[0], tkns[1], tkns[2]
            ebx = f'{bc}:E:{cl}'
            out[rid] = ebx
    return out


def get_aln_map(sam_file):
    if not sam_file:
        return None
    out = {}
    with open(sam_file) as sf:
        for tkns in (line.strip().split() for line in sf):
            out[tkns[0]] = tkns[2]
    return out

def annotate_rid_line(line, aln_map, bc_map):
    line = line.strip()
    rid = line.split()[0][1:]
    if aln_map:
        try:
            aln = 'AL:' + aln_map[rid]
        except KeyError:
            aln = 'AL:UNAL'
        line += ' ' + aln
    try:
        line += ' ' + bc_map[rid]
    except KeyError:
        pass
    line += '\n'
    return line


@click.command()
@click.option('-s', '--sam-file', default=None)
@click.argument('bc_tbl')
def main(sam_file, bc_tbl):
    aln_map, bc_map = get_aln_map(sam_file), get_bc_map(bc_tbl)
    for i, line in enumerate(stdin):
        if (i % 4) == 0:
            stdout.write(annotate_rid_line(line, aln_map, bc_map))
        else:
            stdout.write(line)
        

if __name__ == '__main__':
    main()



import sys


class ProgressBar:

    def __init__(self, total, filelike=sys.stderr, length=100, writeEvery=10):
        self.total = total
        self.events = 0.0
        self.filelike = filelike
        self.length = length
        self.writeEvery = writeEvery
        

        
    def write(self):
        prefix = '{:,} / {:,} '
        prefLength = len(prefix.format(self.total, self.total))
        prefix = prefix.format(self.events, self.total)
        prefix += ' '*(prefLength - len(prefix))
        
        lengthLeft = self.length - prefLength
        lengthLeft -= 2
        p = self.events / self.total
        filled = int(p * lengthLeft)
        bar = '|' + '#'*filled + '-'*(lengthLeft-filled) + '|'
        out = '\r' + prefix + bar
        self.filelike.write(out)

    def finish(self):
        self.write()
        self.filelike.write('\n')

    def increment(self):
        self.events += 1
        if (self.events % self.writeEvery) == 0:
            self.write()
        if self.events == self.total:
            self.finish()
        



import sys

################################################################################

class Fastx:

    __slots__ = ('tags', 'sid', 'seq')
    
    def __init__(self):
        pass
    
    def parseIdLine(self,rsid):
        sid = rsid.strip()
        if sid[0] in ['>', '@']:
            sid = sid[1:]
        sid = sid.split()
        try:
            self.tags = sid[1:]
        except IndexError as ie:
            self.tags = []
        try:
            self.sid = sid[0]
        except IndexError as ie:
            sys.stderr.write('IDLINE: {}\n'.format(rsid))
            raise ie
            
    def __len__(self):
        return len(self.seq)
        
class Fasta( Fastx):

    __slots__ = [] # I am not clear if this is necessary 
    
    def __init__(self, sid, seq):
        super(Fasta, self).__init__()
        self.parseIdLine(sid)
        self.seq = seq.strip()

    def __str__(self):
        tags = '\t'.join( self.tags)
        out = '>{} {}\n{}\n'.format(self.sid, tags, self.seq)
        return out

    @classmethod
    def fromRaw(cls, inp):
        if type(inp) == str:
            inp = inp.split('\n')
        assert len(inp) == 2
        return Fasta( inp[0], inp[1])
    
class Fastq( Fastx):

    __slots__ = ('delim', 'qual')
    
    def __init__(self, sid, seq, delim, qual):
        self.parseIdLine(sid)
        self.seq = seq.strip()
        self.delim = delim.strip()
        self.qual = qual.strip()

        assert len(self.qual) == len(self.seq)
        
    def __str__(self):
        tags = '\t'.join(self.tags)
        out = '@{}\t{}\n{}\n{}\n{}'.format(self.sid,
                                           tags,
                                           self.seq,
                                           self.delim,
                                           self.qual)
        return out

    @classmethod
    def fromRaw(cls, inp):
        if type(inp) == str:
            inp = inp.split('\n')
        assert len(inp) == 4
        return Fastq( inp[0], inp[1], inp[2], inp[3])

    
class ReadPair:

    __slots__ = ('r1', 'r2', 'sid')
    
    def __init__(self, r1, r2):
        assert r1.sid == r2.sid
        assert type(r1) == type(r2)
        self.sid = r1.sid
        self.r1 = r1
        self.r2 = r2

    def __len__(self):
        return len(self.r1) + len(self.r2)

    def __str__(self):
        return str(self.r1) + '\n' + str(self.r2)
        
def iterChunks(filelike, n):
    chunk = []
    for line in filelike:
        chunk.append(line)
        if len(chunk) == n:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


        
################################################################################
        
def iterFastq(filelike, interleaved=False):
    n = 4
    if interleaved:
        n = 8
    for chunk in iterChunks(filelike, n):
        r1 = Fastq.fromRaw( chunk[:4])
        if interleaved:
            r2 = Fastq.fromRaw( chunk[4:])
            yield ReadPair(r1,r2)
        else:
            yield r1


def iterFasta(filelike, interleaved=False):
    n = 2
    if interleaved:
        n = 4
    for chunk in iterChunks(filelike, n):
        r1 = Fasta.fromRaw( chunk[:2])
        if interleaved:
            r2 = Fasta.fromRaw( chunk[2:])
            yield ReadPair(r1,r2)
        else:
            yield r1
            
            



import setuptools

requirements = [
    'scikit-learn',
    'numpy',
    'pandas',
    'scipy',
    'click',
    'argparse',
]


setuptools.setup(
    name="minerva_deconvolve",
    version="1.0.0",
    url="https://github.com/dcdanko/minerva_barcode_deconvolution",
    author="David C. Danko",
    author_email="dcd3001@med.cornell.edu",
    description="Algorithm for deconvolving and clustering barcoded short reads",
    packages=setuptools.find_packages(),
    package_dir={'minerva': 'minerva'},
    ext_modules=[
        setuptools.Extension("cseqs", ["cext_minerva/_cseqs.c", "cext_minerva/cseqs.c"])
    ],
    entry_points={
        'console_scripts': [
            'minerva_deconvolve=minerva.deconvolution.deconvolve_barcodes:main',
            'minerva_enhance_kraken=minerva.kraken.enhance_kraken:main',
            'minerva_eval=minerva.eval.eval_deconvolution:main',
            'minerva_annotate=minerva.eval.annotate_fastq:main',
        ]
    },
    install_requires=requirements,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)






################################################################################

def baseToInt(base):
    if base == 'A':
        return 0
    elif base == 'C':
        return 1
    elif base == 'G':
        return 2
    elif base == 'T':
        return 3
    else:
        return 4

def rcBase(base):
    if base == 'A':
        return 'T'
    elif base =='C':
        return 'G'
    elif base == 'G':
        return 'C'
    elif base == 'T':
        return 'A'
    else:
        return 'N'


def reverseComplement(kmer):
    rc = ''
    for base in kmer[::-1]:
        rc += rcBase( base)
    return rc
        
def canonical( kmer):
    rc = ''
    for base in kmer[::-1]:
        rc += rcBase( base)
        if kmer < rc:
            return kmer
    return rc



import sys
import argparse as ap
import numpy as np
import math
from ..gimmebio.kmers import *


################################################################################
    
def passesFilter(nHits, possibleHits, minHits, maxHits):
    return belowFilter(nHits, possibleHits, maxHits) and aboveFilter(nHits, possibleHits, minHits)

def belowFilter(nHits, possibleHits, maxHits):
    if maxHits >= 1:
        return nHits <= maxHits
    else:
        pHits = nHits / possibleHits
        return pHits <= maxHits
    
def aboveFilter(nHits, possibleHits, minHits):
    if minHits >= 1:
        return nHits >= minHits
    else:
        pHits = nHits / possibleHits
        return pHits >= minHits

